/*
 * Copyright (C) 2005 Texas Instruments.
 *
 * (C) Copyright 2004
 * Jian Zhang, Texas Instruments, jzhang@ti.com.
 *
 * (C) Copyright 2002
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * (C) Copyright 2002
 * Sysgo Real-Time Solutions, GmbH <www.elinos.com>
 * Marius Groeger <mgroeger@sysgo.de>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <asm/arch/mem.h>
#include "image.h"
#include <setup.h>
#include <asm/system.h>

extern int misc_init_r (void);
extern u32 get_mem_type(void);

#ifdef CFG_PRINTF
int print_info(void)
{
	printf("\n\nSauris GmbH SauLoader 1.01 ("
			__DATE__ " - " __TIME__ ")\n");
	return 0;
}
#endif

typedef int (init_fnc_t) (void);

init_fnc_t *init_sequence[] = {
	cpu_init,		/* basic cpu dependent setup */
	board_init,		/* basic board dependent setup */
#ifdef CFG_PRINTF
 	serial_init,		/* serial communications setup */
	print_info,
#endif
  	nand_init,		/* board specific nand init */
  	NULL,
};

static int dram_size;

#define PATCH_ENDIANESS(x) x = ( \
	(*((u8*)(&(x))) << 24) | \
	(*(((u8*)(&(x)))+1) << 16) | \
	(*(((u8*)(&(x)))+2) << 8) | \
	*(((u8*)(&(x)))+3) )

#define PAGE_SIZE 2048

void	(*theKernel)(int zero, int arch, uint params);

struct tag *params;

/*
   WARNING!!! console=ttyO2 must be first, it may directly 
   patched to ttyS2 for old kernel !!!!!!
*/

static char * kernel_cmdline = \
	"console=ttyO2,115200n8 " \
	"ubi.mtd=4 " \
	"root=ubi0:root_volume " \
	"rw " \
	"rootfstype=ubifs " \
	"rootflags=compr=none,bulk_read,no_chk_data_crc " \
	"lpj=2334720 " \
	"ip=off " \
	"printk.time=1";

static void * memmove(void * dest,const void *src,size_t count)
{
	char *tmp, *s;

	if (dest <= src) {
		tmp = (char *) dest;
		s = (char *) src;
		while (count--)
			*tmp++ = *s++;
		}
	else {
		tmp = (char *) dest + count;
		s = (char *) src + count;
		while (count--)
			*--tmp = *--s;
		}

	return dest;
}

static char * strcpy(char * dest,const char *src)
{
	char *tmp = dest;

	while ((*dest++ = *src++) != '\0')
		/* nothing */;
	return tmp;
}

static size_t strlen(const char * s)
{
	const char *sc;

	for (sc = s; *sc != '\0'; ++sc)
		/* nothing */;
	return sc - s;
}


static void setup_start_tag (void)
{
	params = (struct tag *) 0x80000100;

	params->hdr.tag = ATAG_CORE;
	params->hdr.size = tag_size (tag_core);

	params->u.core.flags = 0;
	params->u.core.pagesize = 0;
	params->u.core.rootdev = 0;

	params = tag_next (params);
}

void setup_revision_tag (void)
{

	params->hdr.tag = ATAG_REVISION;
	params->hdr.size = tag_size (tag_revision);
	params->u.revision.rev = 0x20;
	params = tag_next (params);
}

static void setup_memory_tags (void)
{

	params->hdr.tag = ATAG_MEM;
	params->hdr.size = tag_size (tag_mem32);

	params->u.mem.start = 0x80000000;
	params->u.mem.size = dram_size;

	params = tag_next (params);

	params->hdr.tag = ATAG_MEM;
	params->hdr.size = tag_size (tag_mem32);

	params->u.mem.start = 0x80000000;
	params->u.mem.size = 0;

	params = tag_next (params);

}

static void setup_commandline_tag (char *commandline)
{

	params->hdr.tag = ATAG_CMDLINE;
	params->hdr.size =
		(sizeof (struct tag_header) + strlen (commandline) + 1 + 4) >> 2;

	strcpy (params->u.cmdline.cmdline, commandline);

	params = tag_next (params);
}

static void setup_end_tag (void)
{
	params->hdr.tag = ATAG_NONE;
	params->hdr.size = 0;
}

static void cp_delay (void)
{
	volatile int i;

	/* copro seems to need some delay between reading and writing */
	for (i = 0; i < 100; i++)
		nop();
}

/* cache_bit must be either CR_I or CR_C */
static void cache_enable(uint32_t cache_bit)
{
	uint32_t reg;

	reg = get_cr();	/* get control reg. */
	cp_delay();
	set_cr(reg | cache_bit);
}

/* cache_bit must be either CR_I or CR_C */
static void cache_disable(uint32_t cache_bit)
{
	uint32_t reg;

	reg = get_cr();
	cp_delay();
	set_cr(reg & ~cache_bit);
}

void icache_enable(void)
{
	cache_enable(CR_I);
}

void icache_disable(void)
{
	cache_disable(CR_I);
}

void dcache_enable(void)
{
	cache_enable(CR_C);
}

void dcache_disable(void)
{
	cache_disable(CR_C);
}


int disable_interrupts (void)
{
	unsigned long old,temp;
	__asm__ __volatile__("mrs %0, cpsr\n"
			     "orr %1, %0, #0xc0\n"
			     "msr cpsr_c, %1"
			     : "=r" (old), "=r" (temp)
			     :
			     : "memory");
	return (old & 0x80) == 0;
}

static void cache_flush(void)
{
	asm ("mcr p15, 0, %0, c7, c5, 0": :"r" (0));
}

void invalidate_dcache(u32);
void l2_cache_enable(void);
void l2_cache_disable(void);

void start_armboot (void)
{
  	init_fnc_t **init_fnc_ptr;
 	int i, rc;
	uchar *buf;

   	for (init_fnc_ptr = init_sequence; *init_fnc_ptr; ++init_fnc_ptr) {
		if ((*init_fnc_ptr)() != 0) {
			hang ();
		}
	}

	misc_init_r();

	dram_size = mem_test();

	if ( dram_size < 0)
		hang();

	buf =  (uchar*) CFG_LOADADDR;

	if ((get_mem_type() == MMC_ONENAND) || (get_mem_type() == MMC_NAND)) {
		printf("Loading U-Boot from MMC\n");
		buf += mmc_boot(buf);
	}
	
	if (buf == (uchar*)CFG_LOADADDR) {

		if (get_mem_type() == GPMC_NAND) {
			u32 nand_addr;
			
			printf("Attempting to load kernel directly from NAND\n");
			nand_addr = 0x280000;
			if (!(rc=nand_read_page(buf, nand_addr))) {
			
				struct image_header *ih = (struct image_header *) buf;
				u32 load_addr;
				int size;
				void (*theKernel)(int zero, int arch, uint params);

				PATCH_ENDIANESS(ih->ih_magic);
				PATCH_ENDIANESS(ih->ih_hcrc);
				PATCH_ENDIANESS(ih->ih_time);
				PATCH_ENDIANESS(ih->ih_size);
				PATCH_ENDIANESS(ih->ih_load);
				PATCH_ENDIANESS(ih->ih_ep);
				PATCH_ENDIANESS(ih->ih_dcrc);
				
				if (ih->ih_magic == IH_MAGIC &&
				    ih->ih_os == IH_OS_LINUX &&
				    ih->ih_arch == IH_ARCH_ARM &&
				    ih->ih_type == IH_TYPE_KERNEL &&
				    ih->ih_comp == IH_COMP_NONE) {

					// patch kernel_cmdline ttyO2->ttyS2 for 2.6.32
					if (ih->ih_name[10] == '3' && ih->ih_name[11] == '2')
						kernel_cmdline[11] = 'S';

					printf("Kernel image \"%s\" found\n", ih->ih_name);
					printf("Load address = 0x%08X\n", ih->ih_load);
					printf("Entry point  = 0x%08X\n", ih->ih_ep);
					printf("Image size   = %d\n", ih->ih_size);
					printf("Loading kernel image ...");

					load_addr = ih->ih_load;
					theKernel = (void (*)(int, int, uint)) ih->ih_ep;
					size = ih->ih_size;

					memmove((uchar*)(load_addr), 
						buf+sizeof(struct image_header), 
						PAGE_SIZE - sizeof(struct image_header) );

					load_addr += PAGE_SIZE - sizeof(struct image_header);
					for (i = PAGE_SIZE - sizeof(struct image_header); i < size; i+=PAGE_SIZE) {
						nand_addr += PAGE_SIZE;
						rc=nand_read_page((u8*)load_addr, nand_addr);
						if (rc) break;
						load_addr += PAGE_SIZE;
					}

					if (!rc) {

						printf(" ok\nStarting kernel \n\n");

						setup_start_tag();
						setup_revision_tag();
						setup_memory_tags();
						setup_commandline_tag(kernel_cmdline);
						setup_end_tag();

						disable_interrupts();
						icache_disable();
						dcache_disable();

						cache_flush();
						l2_cache_disable();
						invalidate_dcache(get_device_type());
						i = 0;
						/* mem barrier to sync up things */
						asm("mcr p15, 0, %0, c7, c10, 4": :"r"(i));
						l2_cache_enable();

						theKernel(0, 2200, 0x80000100);

					} else printf(" failed \n");
				}
			}

			printf("Loading U-Boot from NAND\n");
			for (i = NAND_UBOOT_START; i < NAND_UBOOT_END; i+= NAND_BLOCK_SIZE) {
				if (!(rc=nand_read_block(buf, i)))
					buf += NAND_BLOCK_SIZE; /* advance buf ptr */
				else 
					break;
			}
			printf("%ld bytes read from NAND\n", buf - (uchar*)CFG_LOADADDR);
		}
	}

	if (buf == (uchar *)CFG_LOADADDR)
		hang();

	/* go run U-Boot and never return */
  	printf("Starting OS Bootloader...\n");
 	((init_fnc_t *)CFG_LOADADDR)();

	/* should never come here */
}

void hang (void)
{
	/* call board specific hang function */
	board_hang();

	/* if board_hang() returns, hange here */
	printf("X-Loader hangs\n");
	for (;;);
}
